from ast import Call
from typing import Callable, Optional, Union
from .abstractions.objective_function import ObjectiveFunction, OptimizationType
from ..modeling.abstractions.model import Model, TInput, TTarget
from ..evaluation.abstractions.evaluation_metric import EvaluationMetric, TModel, EvaluationContext


class CustomObjectiveFunction(ObjectiveFunction[TInput, TTarget, TModel]):
    def __init__(self, expression: Union[Callable[[EvaluationContext[TInput, TTarget, TModel]], float], EvaluationMetric[TInput, TTarget, TModel]], optimization_type: Optional[OptimizationType] = OptimizationType.MAX):
        if expression is None:
            raise ValueError("expression can't be empty")

        self.expression: Union[Callable[[EvaluationContext[TInput, TTarget, TModel]], float], EvaluationMetric[TInput, TTarget, TModel]] = expression

        self.__optimization_type: OptimizationType = optimization_type

    @property
    def optimization_type(self) -> OptimizationType:
        return self.__optimization_type

    def calculate_score(self, context: EvaluationContext[TInput, TTarget, TModel]) -> float:
        return self.expression(context)


from logging import Logger
from typing import List
from ..modeling.abstractions.model import Model, TInput, TTarget
from .abstractions.evaluation_metric import EvaluationContext, EvaluationMetric, Prediction, TModel

def default_evaluation(logger: Logger, evaluation_context: EvaluationContext[TInput, TTarget, TModel], model: TModel, input_batch: List[TInput], target_batch: List[TTarget]) -> List[TTarget]:
    return model.predict_batch(input_batch=input_batch)


from abc import ABC
from dataclasses import dataclass
from typing import Any, Dict, Generic, List, TypeVar
from experimentation.experiment.abstractions.experiment_settings import ExperimentSettings
from experimentation.experiment.abstractions.experiment_result import ExperimentResult
from experimentation.experiment.abstractions.experimentation_service import ExperimentationService

from ...modeling.abstractions.model import TInput, TTarget
from ...evaluation.abstractions.evaluation_metric import TModel, EvaluationContext

@dataclass
class InstanceSettings:
    name: str
    params: Dict[str, Any]

@dataclass
class MachineLearningExperimentResult(Generic[TModel], ExperimentResult):
    model: TModel
    scores: Dict[str, float]

@dataclass    
class MachineLearningExperimentSettings(ExperimentSettings):    
    model_settings: List[InstanceSettings]
    training_service_settings: List[InstanceSettings]
    evaluation_service_settings: List[InstanceSettings]
    evaluation_dataset_settings: List[Dict[str, InstanceSettings]]
    training_dataset_settings: List[Dict[str, InstanceSettings]]
    evaluation_metric_settings: List[Dict[str, InstanceSettings]]
    objective_function_settings: List[Dict[str, InstanceSettings]]
    stop_condition_settings: List[Dict[str, InstanceSettings]]

@dataclass    
class MachineLearningRunSettings():    
    experiment_name: str
    model_settings: InstanceSettings
    training_service_settings: InstanceSettings
    evaluation_service_settings: InstanceSettings
    evaluation_dataset_settings: Dict[str, InstanceSettings]
    training_dataset_settings: Dict[str, InstanceSettings]
    evaluation_metric_settings: Dict[str, InstanceSettings]
    objective_function_settings: Dict[str, InstanceSettings]
    stop_condition_settings: Dict[str, InstanceSettings]

@dataclass 
class MachineLearningRunResult(Generic[TModel]):
    run_settings: MachineLearningRunSettings
    model: TModel
    scores: Dict[str, float]

@dataclass
class MachineLearningExperimentResult(Generic[TModel], ExperimentResult):
    run_results: List[MachineLearningRunResult[TModel]]

class MachineLearningExperimentationService(Generic[TModel], 
ExperimentationService[MachineLearningExperimentSettings, MachineLearningExperimentResult[TModel]]):
    pass


from typing import Any, Callable, Dict, Generic, List, Type, Union

from .abstractions.instance_factory import InstanceFactory, TInstance
from .abstractions.machine_learning_experimentation_service import InstanceSettings
from .default_instance_factory import DefaultInstanceFactory

class DefaultDictInstanceFactory(Generic[TInstance], InstanceFactory[Dict[str, InstanceSettings], Dict[str, TInstance]]):
    def __init__(self, available_types: Dict[str, Union[Type[TInstance], Callable[[Dict[str, Any]], TInstance]]]):
        if available_types is None:
            raise ValueError("available_types")

        self.__single_instance_factory: InstanceFactory[InstanceSettings, TInstance] = DefaultInstanceFactory[TInstance](available_types)

    def create(self, settings: Dict[str, InstanceSettings]) -> Dict[str, TInstance]:
        if settings is None:
            raise ValueError("settings")

        instances: Dict[str, TInstance] = {}

        for alias, instance_item in settings.items():
            instances[alias] = self.__single_instance_factory.create(instance_item)

        return instances

        


from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, Generic, List, Tuple, Union
from torch.utils.data.dataset import Dataset

from .model_factory import ModelFactory

from .objective_function import ObjectiveFunction

from ...modeling.abstractions.model import TInput, TTarget
from ...evaluation.abstractions.evaluation_metric import TModel

class ParameterTuningService(Generic[TInput, TTarget, TModel], ABC):

    @abstractmethod
    async def search(self, model_factory: ModelFactory[TModel], params: Dict[str, List[Any]], dataset: Dataset[Tuple[TInput, TTarget]], objective_functions: Dict[str, ObjectiveFunction[TInput, TTarget, TModel]], primary_objective: str) -> Dict[str, Any]:
        pass


from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, Generic, List, Optional, Tuple, Union
from numpy import ndarray
from sklearn.model_selection import GridSearchCV
from torch.utils.data.dataset import Dataset
from sklearn import metrics

from ..modeling.adapter.sklearn_estimator_adapter import SkleanEstimatorAdapter
from .abstractions.model_factory import ModelFactory
from .abstractions.objective_function import ObjectiveFunction, OptimizationType
from ..modeling.abstractions.model import TInput, TTarget, Model
from ..evaluation.abstractions.evaluation_metric import Prediction, TModel, EvaluationContext
from .abstractions.parameter_tuning_service import ParameterTuningService

def score_function(estimator: SkleanEstimatorAdapter[Model[TInput, TTarget]], y_target: ndarray, y_predicted: ndarray, objective_function: ObjectiveFunction[TInput, TTarget, Model[TInput, TTarget]]) -> float:
    
    predictions: List[Prediction] = [Prediction(input=None, prediction=predicted, target=target) for target, predicted in zip(y_target, y_predicted)]
    context: EvaluationContext[TInput, TTarget, Model[TInput, TTarget]] = EvaluationContext[TInput, TTarget, Model[TInput, TTarget]](model=None, predictions=predictions)

    return objective_function.calculate_score(context=context)

class GridSearchTuningService(ParameterTuningService[TInput, TTarget, Model[TInput, TTarget]]):
    def __init__(self, folds: int = 5):
        self.__folds: int = folds

    async def search(self, model_factory: ModelFactory[TModel], params: Dict[str, List[Any]], dataset: Dataset[Tuple[TInput, TTarget]], objective_functions: Dict[str, ObjectiveFunction[TInput, TTarget, Model[TInput, TTarget]]], primary_objective: str) -> Dict[str, Any]:
        if model_factory is None:
            raise ValueError("model_factory can't be empty")

        if params is None:
            raise ValueError("params can't be empty")

        if dataset is None:
            raise ValueError("dataset can't be empty")

        if objective_functions is None:
            raise ValueError("objective_functions can't be empty")

        factory_params: Dict[str, Any] = {key: values[0] for key, values in params.items()}
        model: TModel = model_factory.create(factory_params)
        estimator = SkleanEstimatorAdapter[Model[TInput, TTarget]](model)
        scorer = {key: metrics.make_scorer(lambda y, y_pred, **kwargs: score_function(model, y, y_pred, objective_function), greater_is_better=objective_function.optimization_type == OptimizationType.MAX) for key, objective_function in objective_functions.items()}

        inputs: List[TInput] = [value[0] for value in dataset]
        targets: List[TTarget] = [value[1] for value in dataset]

        grid_search = GridSearchCV(estimator=estimator, param_grid=params, cv=self.__folds, scoring=scorer, refit=primary_objective)
        grid_search.fit(X=inputs, y=targets)

        return grid_search.best_params_




import unittest
from unittest.mock import Mock, PropertyMock, patch
from typing import Any, Coroutine, List, Dict, Tuple
from faker import Faker

from ..evaluation.custom_evaluation_metric import CustomEvaluationMetric
from ..evaluation.abstractions.evaluation_metric import EvaluationContext, Prediction, TTarget, TModel, TInput
from ..training.abstractions.stop_condition import TrainingContext, StopCondition
from ..training.custom_stop_condition import CustomStopCondition
from ..modeling.abstractions.model import Model

class CustomStopConditionTestCase(unittest.TestCase):
    def setUp(self):
        fake = Faker()

        self.training_context_patcher = patch('machine_learning.training.abstractions.stop_condition.TrainingContext', new=TrainingContext[float, float, Model[float, float]])

        self.training_context: TrainingContext[float, float, Model[float, float]] = self.training_context_patcher.start()

    def tearDown(self):
        self.training_context_patcher.stop()

    def test_is_satisfied_should_return_true(self):
        stop_condition: StopCondition[TInput, TTarget, Model[float, float]] = CustomStopCondition[float, float, Model[float, float]](expression=lambda x: True)

        assert stop_condition.is_satisfied(self.training_context)


from abc import ABC, abstractmethod
from typing import Any, Dict, Generic

from ...evaluation.abstractions.evaluation_metric import TModel

class ModelFactory(Generic[TModel], ABC):

    @abstractmethod
    def create(self, params: Dict[str, Any]) -> TModel:
        pass


from abc import ABC, abstractmethod
from typing import TypeVar, List, Generic
from dataclasses import dataclass
from ...modeling.abstractions.model import Model, TInput, TTarget

TModel = TypeVar('TModel', bound=Model)

@dataclass(frozen=True)
class Prediction(Generic[TInput, TTarget]):
    input: TInput
    prediction: TTarget
    target: TTarget

@dataclass(frozen=True)
class EvaluationContext(Generic[TInput, TTarget, TModel], ABC):
    model: TModel
    dataset_name: str
    predictions: List[Prediction[TInput, TTarget]]

class EvaluationMetric(Generic[TInput, TTarget, TModel], ABC):
    
    @abstractmethod
    def calculate_score(self, context: EvaluationContext[TInput, TTarget, TModel]) -> float:
        pass

    def __call__(self, context: EvaluationContext[TInput, TTarget, TModel]) -> float:
        return self.calculate_score(context)


import unittest
from unittest.mock import Mock, PropertyMock, patch
from faker import Faker

from ..parameter_tuning.custom_model_factory import CustomModelFactory
from ..modeling.abstractions.model import Model

class CustomModelFactoryTestCase(unittest.TestCase):
    def setUp(self):
        fake = Faker()

        self.model_patcher = patch('machine_learning.modeling.abstractions.model', new=Model[float, float])

        self.model: Model[float, float] = self.model_patcher.start()

    def tearDown(self):
        self.model_patcher.stop()

    def test_create_valid_params_should_return_model_instance(self):
        model_factory: CustomModelFactory[Model[float, float]] = CustomModelFactory[Model[float, float]](expression=lambda params: self.model)

        model = model_factory.create({})

        assert not model is None


from ast import Call
import asyncio
from audioop import mul
import itertools
from logging import Logger
import logging
from multiprocessing.dummy import Pool
from optparse import Option
from typing import Any, Callable, Coroutine, Dict, Generic, List, Optional, Tuple, TypeVar, Union, final
from uuid import UUID
import uuid
from py import process
from torch.utils.data import Dataset, random_split
import nest_asyncio
import dill
import pathos.multiprocessing as multiprocessing
from dataclasses import dataclass
from pytorch_lightning.loggers.base import LightningLoggerBase

from .abstractions.instance_factory import InstanceFactory
from ..evaluation.abstractions.evaluation_metric import EvaluationMetric, EvaluationContext, TModel
from ..training.abstractions.stop_condition import StopCondition, TrainingContext
from ..parameter_tuning.abstractions.objective_function import ObjectiveFunction
from ..modeling.abstractions.model import TInput, TTarget
from ..evaluation.abstractions.evaluation_service import EvaluationService, Score
from ..training.abstractions.training_service import TrainingService
from .abstractions.machine_learning_experimentation_service import MachineLearningExperimentationService, MachineLearningExperimentSettings, MachineLearningExperimentResult, MachineLearningRunResult, MachineLearningRunSettings, InstanceSettings
from .default_instance_factory import DefaultInstanceFactory
from .default_dict_instance_factory import DefaultDictInstanceFactory

nest_asyncio.apply()

TSettings = TypeVar('TSettings', InstanceSettings, Dict[str, InstanceSettings])
TInstance = TypeVar('TInstance')
TRunMetricLogger = TypeVar('TRunMetricLogger', bound=LightningLoggerBase)

FactoryAlias = Union[InstanceFactory[TSettings, TInstance], Callable[[TSettings], TInstance]]
ModelFactoryAlias = FactoryAlias[InstanceSettings, TModel]
TrainingServiceFactoryAlias = FactoryAlias[InstanceSettings, TrainingService[TInput, TTarget, TModel]]
EvaluationServiceFactoryAlias = FactoryAlias[InstanceSettings, EvaluationService[TInput, TTarget, TModel]]
TrainingDatasetFactoryAlias = FactoryAlias[Dict[str, InstanceSettings], Dict[str, Dataset[Tuple[TInput, TTarget]]]]
EvaluationDatasetFactoryAlias = FactoryAlias[Dict[str, InstanceSettings], Dict[str, Dataset[Tuple[TInput, TTarget]]]]
EvaluationMetricFactoryAlias = FactoryAlias[Dict[str, InstanceSettings], Dict[str, EvaluationMetric[TInput, TTarget, TModel]]]
ObjectiveFunctionFactoryAlias = FactoryAlias[Dict[str, InstanceSettings], Dict[str, ObjectiveFunction[TInput, TTarget, TModel]]]
StopConditionFactoryAlias = FactoryAlias[Dict[str, InstanceSettings], Dict[str, StopCondition[TInput, TTarget, TModel]]]

@dataclass
class RunCheckpoint:
    id: UUID
    run_settings: MachineLearningRunSettings
    train_run_id: UUID

@dataclass
class ExperimentCheckpoint:
    id: UUID
    name: str
    experiment_settings: MachineLearningExperimentSettings
    runs: Dict[UUID, MachineLearningRunSettings]

class DefaultMachineLearningExperimentationService(Generic[TRunMetricLogger], MachineLearningExperimentationService[TModel]):
    def __init__(self,
    model_factory: ModelFactoryAlias[TModel], 
    training_service_factory: TrainingServiceFactoryAlias[TInput, TTarget, TModel],
    evaluation_service_factory: EvaluationServiceFactoryAlias[TInput, TTarget, TModel],
    training_dataset_factory: TrainingDatasetFactoryAlias[TInput, TTarget], 
    test_dataset_factory: EvaluationDatasetFactoryAlias[TInput, TTarget],
    evaluation_metric_factory: EvaluationMetricFactoryAlias[TInput, TTarget, TModel],
    objective_function_factory: ObjectiveFunctionFactoryAlias[TInput, TTarget, TModel],
    stop_condition_factory: StopConditionFactoryAlias[TInput, TTarget, TModel],
    run_metric_logger_factory: Callable[[str, UUID, UUID, MachineLearningRunSettings, bool], LightningLoggerBase],
    event_loop: Optional[asyncio.AbstractEventLoop] = None,
    logger: Optional[Logger] = None, 
    process_pool: Optional[multiprocessing.ProcessPool] = None,
    run_logger_factory: Optional[Callable[[str, UUID, UUID, MachineLearningRunSettings, bool], Logger]] = None,
    experiment_logger_factory: Optional[Callable[[str, UUID, MachineLearningExperimentSettings, bool], Logger]] = None,
    save_run_checkpoint_hook: Optional[Callable[[Logger, RunCheckpoint], None]] = None,
    load_run_checkpoint_hook: Optional[Callable[[Logger, UUID], Optional[RunCheckpoint]]] = None,
    save_experiment_checkpoint_hook: Optional[Callable[[Logger, ExperimentCheckpoint], None]] = None,
    load_experiment_checkpoint_hook: Optional[Callable[[Logger, str, UUID], Optional[ExperimentCheckpoint]]] = None):

        if model_factory is None:
            raise ValueError("model_factory")

        if training_service_factory is None:
            raise ValueError("training_service_factory")

        if evaluation_service_factory is None:
            raise ValueError("evaluation_service_factory")

        if training_dataset_factory is None:
            raise ValueError("training_dataset_factory")

        if test_dataset_factory is None:
            raise ValueError("test_dataset_factory")        
            
        if evaluation_metric_factory is None:
            raise ValueError("evaluation_metric_factory")

        if objective_function_factory is None:
            raise ValueError("objective_function_factory")

        if stop_condition_factory is None:
            raise ValueError("stop_condition_factory")

        if run_metric_logger_factory is None:
            raise ValueError("run_metric_logger_factory")

        self.__logger = logger if not logger is None else logging.getLogger()
        self.__pool: multiprocessing.ProcessPool = process_pool if not process_pool is None else multiprocessing.ProcessPool(4)
        self.__event_loop: asyncio.AbstractEventLoop = event_loop if not event_loop is None else asyncio.get_event_loop()
        self.__run_logger_factory: Callable[[str, UUID, UUID, MachineLearningRunSettings, bool], Logger] = run_logger_factory if not run_logger_factory is None else lambda experiment_name, experiment_id, run_id, settings, continue_run: logging.getLogger(str(run_id))
        self.__run_metric_logger_factory: Optional[Callable[[str, UUID, UUID, MachineLearningRunSettings, bool], LightningLoggerBase]] = run_metric_logger_factory
        self.__experiment_logger_factory: Callable[[str, UUID, MachineLearningExperimentSettings, bool], Logger] = experiment_logger_factory if not experiment_logger_factory is None else lambda experiment_name, uuid, settings, continue_run: logging.getLogger(str(uuid))
        self.__model_factory: ModelFactoryAlias[TModel] = model_factory
        self.__training_service_factory: TrainingServiceFactoryAlias[TInput, TTarget, TModel] = training_service_factory
        self.__evaluation_service_factory: EvaluationServiceFactoryAlias[TInput, TTarget, TModel] = evaluation_service_factory
        self.__training_dataset_factory: TrainingDatasetFactoryAlias[TInput, TTarget] = training_dataset_factory
        self.__test_dataset_factory: EvaluationDatasetFactoryAlias[TInput, TTarget] = test_dataset_factory
        self.__evaluation_metric_factory: EvaluationMetricFactoryAlias[TInput, TTarget, TModel] = evaluation_metric_factory
        self.__objective_function_factory: ObjectiveFunctionFactoryAlias[TInput, TTarget, TModel] = objective_function_factory
        self.__stop_condition_factory: StopConditionFactoryAlias[TModel] = stop_condition_factory

        self.__save_run_checkpoint_hook: Optional[Callable[[Logger, RunCheckpoint], None]] = save_run_checkpoint_hook
        self.__load_run_checkpoint_hook: Optional[Callable[[Logger, UUID], Optional[RunCheckpoint]]] = load_run_checkpoint_hook

        self.__save_experiment_checkpoint_hook: Optional[Callable[[Logger, ExperimentCheckpoint], None]] = save_experiment_checkpoint_hook
        self.__load_experiment_checkpoint_hook: Optional[Callable[[Logger, str, UUID], Optional[ExperimentCheckpoint]]] = load_experiment_checkpoint_hook

    def __getstate__(self):
        self_dict = self.__dict__.copy()
        del self_dict['_DefaultMachineLearningExperimentationService__pool']
        del self_dict['_DefaultMachineLearningExperimentationService__event_loop']
        return self_dict

    def __setstate__(self, state):
        self.__dict__.update(state)

    def __save_run_checkpoint(self, logger: Logger, checkpoint: RunCheckpoint):
        if not self.__save_run_checkpoint_hook is None:
            logger.info("creating run checkpoint...")

            self.__save_run_checkpoint_hook(logger, checkpoint)

            logger.info("run checkpoint created.")

    def __load_run_checkpoint(self, logger: Logger, id: UUID) -> RunCheckpoint:
        if self.__load_run_checkpoint_hook is None:
            return None

        logger.info("Loading last run checkpoint...")

        return self.__load_run_checkpoint_hook(logger, id)

    def __save_experiment_checkpoint(self, logger: Logger, checkpoint: ExperimentCheckpoint):
        if not self.__save_experiment_checkpoint_hook is None:
            logger.info("creating experiment checkpoint...")

            self.__save_experiment_checkpoint_hook(logger, checkpoint)

            logger.info("experiment checkpoint created.")

    def __load_experiment_checkpoint(self, logger: Logger, name: str, id: UUID) -> ExperimentCheckpoint:
        if self.__load_experiment_checkpoint_hook is None:
            return None

        logger.info("Loading last experiment checkpoint...")

        return self.__load_experiment_checkpoint_hook(logger, name, id)

    def execute_run(self, run_settings: MachineLearningRunSettings, experiment_id: Optional[UUID] = None, run_id: Optional[UUID] = None) -> MachineLearningRunResult[TModel]:
        run_id = uuid.uuid4() if run_id is None else run_id

        run_logger: Logger = self.__run_logger_factory(run_settings.experiment_name, experiment_id, run_id, run_settings)
        run_metric_logger: TRunMetricLogger = self.__run_metric_logger_factory(run_settings.experiment_name, experiment_id, run_id, run_settings)

        checkpoint: RunCheckpoint = self.__load_run_checkpoint(run_logger, run_id)

        train_run_id: UUID = None

        if not checkpoint is None:
            train_run_id = checkpoint.train_run_id
            run_settings = checkpoint.run_settings
        else:
            train_run_id = uuid.uuid4()

            new_checkpoint: RunCheckpoint = RunCheckpoint(run_id, run_settings, train_run_id)
            self.__save_run_checkpoint(run_logger, new_checkpoint)

        event_loop: asyncio.AbstractEventLoop = asyncio.get_event_loop()
        result: MachineLearningRunResult[TModel] = None

        try:
            run_logger.info("executing run...")

            model: TModel = self.__model_factory(run_settings.model_settings)
            training_service: TrainingService[TInput, TTarget, TModel] = self.__training_service_factory(run_settings.training_service_settings)
            training_datasets: Dict[str, Dataset[Tuple[TInput, TTarget]]] = self.__training_dataset_factory(run_settings.training_dataset_settings)
            stop_conditions: Dict[str, StopCondition[TInput, TTarget, TModel]] =  self.__stop_condition_factory(run_settings.stop_condition_settings)
            objective_functions: Dict[str, ObjectiveFunction[TInput, TTarget, TModel]] = self.__objective_function_factory(run_settings.objective_function_settings)
            evaluation_service: EvaluationService[TInput, TTarget, TModel] = self.__evaluation_service_factory(run_settings.evaluation_service_settings)
            evaluation_datasets: Dict[str, Dataset[Tuple[TInput, TTarget]]] = self.__test_dataset_factory(run_settings.evaluation_dataset_settings)
            evaluation_metrics: Dict[str, EvaluationMetric[TInput, TTarget, TModel]] = self.__evaluation_metric_factory(run_settings.evaluation_metric_settings)
            
            model = event_loop.run_until_complete(training_service.train_on_multiple_datasets(model=model, datasets=training_datasets, stop_conditions=stop_conditions, objective_functions=objective_functions, id=train_run_id))
            scores: Dict[str, Dict[str, Score]] = event_loop.run_until_complete(evaluation_service.evaluate_on_multiple_datasets(model=model, evaluation_datasets=evaluation_datasets, evaluation_metrics=evaluation_metrics))

            result = MachineLearningRunResult[TModel](run_settings=run_settings, model=model, scores=scores)
        except Exception as ex:
            run_logger.exception(msg=ex, exc_info=True, stack_info=True)
        finally:
            run_logger.info("finished run.")
            logging.shutdown()

        return result

    async def run_experiment(self, experiment_settings: MachineLearningExperimentSettings, experiment_id: Optional[UUID] = None) -> MachineLearningExperimentResult[TModel]:
        experiment_id: UUID = uuid.uuid4() if experiment_id is None else experiment_id
        experiment_logger: Logger = self.__experiment_logger_factory(experiment_settings.name, experiment_id, experiment_settings, False)

        checkpoint: ExperimentCheckpoint = self.__load_experiment_checkpoint(experiment_logger, experiment_settings.name, experiment_id)

        runs: Dict[UUID, MachineLearningRunSettings] = None

        if not checkpoint is None:
            runs = checkpoint.runs
            experiment_settings = checkpoint.experiment_settings
        else:
            combinations: List[Tuple] = itertools.product(experiment_settings.model_settings, experiment_settings.training_service_settings, 
            experiment_settings.evaluation_service_settings, experiment_settings.evaluation_dataset_settings, experiment_settings.training_dataset_settings,  
            experiment_settings.evaluation_metric_settings, experiment_settings.objective_function_settings, experiment_settings.stop_condition_settings)

            runs = {uuid.uuid4(): MachineLearningRunSettings(experiment_settings.name, *combination) for combination in combinations}

            new_checkpoint: ExperimentCheckpoint = ExperimentCheckpoint(experiment_id, experiment_settings.name, experiment_settings, runs)
            self.__save_experiment_checkpoint(experiment_logger, new_checkpoint)

        experiment_logger.info(f"running experiment {experiment_settings.name}...")

        result: MachineLearningExperimentResult[TModel] = None

        try:
            results: List[MachineLearningRunResult[TModel]] = self.__pool.map(self.execute_run, [settings for run_id, settings in runs.items()], [experiment_id for run_id, settings in runs.items()], [run_id for run_id, settings in runs.items()])

            result = MachineLearningExperimentResult[TModel](results)
        except Exception as ex:
            experiment_logger.critical(msg=ex, exc_info=True, stack_info=True)
        finally:
            experiment_logger.info(f"finished experiment {experiment_settings.name}.")

        return result

    async def __run_experiment(self, experiment_settings: Tuple[str, MachineLearningExperimentSettings]) -> Tuple[str, MachineLearningExperimentResult[TModel]]:
        result = await self.run_experiment(experiment_settings[1])

        return (experiment_settings[0], result)

    async def run_experiments(self, experiment_settings: Dict[str, MachineLearningExperimentSettings]) -> Dict[str, MachineLearningExperimentResult[TModel]]:
        self.__logger.info(f"running {len(experiment_settings)} experiments...")
        
        experiment_tasks: List[Coroutine[Any, Any, Tuple[str, MachineLearningExperimentResult[TModel]]]] = [self.__run_experiment(settings) for settings in experiment_settings.items()]

        results: List[Tuple[str, MachineLearningExperimentResult[TModel]]] = await asyncio.gather(*experiment_tasks)

        result:  Dict[str, MachineLearningExperimentResult[TModel]] = dict(results)

        self.__logger.info(f"finished all {len(experiment_settings)} experiments")

        return result


from typing import Any, Callable, Dict, Generic, Type, Union

from .abstractions.instance_factory import InstanceFactory, TInstance
from .abstractions.machine_learning_experimentation_service import InstanceSettings

class DefaultInstanceFactory(Generic[TInstance], InstanceFactory[InstanceSettings, TInstance]):
    def __init__(self, available_types: Dict[str, Union[Type[TInstance], Callable[[Dict[str, Any]], TInstance]]]):
        if available_types is None:
            raise ValueError("available_types")

        self.__available_types: Dict[str, Union[Type[TInstance], Callable[[Dict[str, Any]], TInstance]]] = available_types

    def create(self, settings: InstanceSettings) -> TInstance:
        if settings is None:
            raise ValueError("settings")

        if settings.name is None:
            raise ValueError("settings.name")

        if settings.params is None:
            raise ValueError("settings.params")

        if not settings.name in self.__available_types:
            raise KeyError(f'No matching type with name {settings.name} found.')

        type: Union[Type[TInstance], Callable[[Dict[str, Any]], TInstance]] = self.__available_types[settings.name]

        return type(**settings.params)


import unittest
from unittest.mock import Mock, PropertyMock, patch
from typing import Any, Coroutine, List, Dict, Tuple
from faker import Faker

from ..parameter_tuning.custom_objective_function import CustomObjectiveFunction

from ..evaluation.abstractions.evaluation_metric import EvaluationContext, Prediction, TTarget, TModel, TInput
from ..modeling.abstractions.model import Model

def accuracy(context: EvaluationContext[float, float, Model[float, float]]) -> float:
    predictions: List[float] = [prediction.prediction for prediction in context.predictions]
    targets: List[float] = [prediction.target for prediction in context.predictions]
    number_samples: int = len(predictions)

    return (predictions == targets) / number_samples

class CustomObjectiveFunctionTestCase(unittest.TestCase):
    def setUp(self):
        fake = Faker()

        self.evaluation_context_patcher = patch('machine_learning.evaluation.abstractions.evaluation_metric.EvaluationContext', new=EvaluationContext[float, float, Model[float, float]])

        self.evaluation_context: EvaluationContext[float, float, Model[float, float]] = self.evaluation_context_patcher.start()

        self.evaluation_context.predictions = PropertyMock(return_value=[Prediction[float, float](fake.pyfloat(positive=True), fake.pyfloat(positive=True), fake.pyfloat(positive=True)) for i in range(10)])

    def tearDown(self):
        self.evaluation_context_patcher.stop()

    def test_calculate_score_valid_predictions_should_return_accuracy_score(self):
        accuracy_metric: CustomObjectiveFunction[float, float, Model[float, float]] = CustomObjectiveFunction[float, float, Model[float, float]](expression=accuracy)

        assert accuracy_metric.calculate_score(self.evaluation_context) == accuracy(self.evaluation_context)


from typing import Any, Callable, Dict
from .abstractions.model_factory import ModelFactory
from ..evaluation.abstractions.evaluation_metric import TModel

class CustomModelFactory(ModelFactory[TModel]):
    def __init__(self, expression: Callable[[Dict[str, Any]], TModel]):
        if expression is None:
            raise ValueError("expression can't be empty")

        self.__expression: Callable[[Dict[str, Any]], TModel] = expression

    def create(self, params: Dict[str, Any]) -> TModel:
        return self.__expression(params)


from typing import Callable

from machine_learning.modeling.abstractions.model import TTarget, TInput
from .abstractions.evaluation_metric import EvaluationContext, EvaluationMetric, TModel

class CustomEvaluationMetric(EvaluationMetric[TInput, TTarget, TModel]):
    def __init__(self, expression: Callable[[EvaluationContext[TInput, TTarget, TModel]], float]):
        if expression is None:
            raise ValueError("expression can't be empty")

        self.expression: Callable[[EvaluationContext[TInput, TTarget, TModel]], float] = expression

    def calculate_score(self, context: EvaluationContext[TInput, TTarget, TModel]) -> float:
        return self.expression(context)


from asyncio import coroutine
import asyncio
from dataclasses import dataclass
from re import S
import unittest
from unittest.mock import MagicMock, Mock, PropertyMock, patch
import torch
from torch.utils.data import Dataset
from dataset_handling.dataloader import DataLoader
from typing import Any, Coroutine, List, Dict, Tuple
from faker import Faker
import random

from ..modeling.pytorch_model import PytorchModel

from torch.nn import Module
from torch.optim import Optimizer

class PytorchModelTestCase(unittest.TestCase):
    def setUp(self):
        fake = Faker()

        self.module_patcher = patch('torch.nn.Module', new=Module)
        self.optimizer_patcher = patch('torch.optim.Optimizer', new=Optimizer)
        self.loss_function_patcher = patch('torch.nn.Module', new=Module)

        self.module: Module = self.module_patcher.start()

        self.module.__init__ = Mock(return_value=None)
        self.module.train = Mock()

        self.optimizer: Optimizer = MagicMock(spec=Optimizer)

        self.optimizer.zero_grad = Mock()
        self.optimizer.step = Mock()

        self.loss_function: Module = self.loss_function_patcher.start()

        self.loss_function.__init__ = Mock(return_value=None)
        self.loss_function.backward = Mock()

        self.model: PytorchModel[float, float] = PytorchModel[float, float](pytorch_module=self.module, device=torch.device('cpu'), loss_function=self.loss_function, optimizer=self.optimizer)

        self.inputs: List[float] = [fake.pyfloat(positive=True) for i in range(10)]
        self.targets: List[float] = [fake.pyfloat(positive=True) for i in range(10)]

    def tearDown(self):
        self.module_patcher.stop()
        self.optimizer_patcher.stop()
        self.loss_function_patcher.stop()

    def test_predict_small_dataset_should_set_model_to_predict_and_call_it(self):
        self.model.predict(input=self.inputs[0])

        self.module.train.assert_called()
        self.module.__init__.assert_called()

    def test_predict_batch_small_dataset_should_set_model_to_predict_and_call_it(self):
        self.model.predict_batch(input_batch=self.inputs)

        self.module.train.assert_called()
        self.module.__init__.assert_called()


from asyncio import coroutine
import asyncio
from dataclasses import dataclass
from re import S
import unittest
from unittest.mock import MagicMock, Mock, patch
from torch import isin
from torch.utils.data import Dataset
from dataset_handling.dataloader import DataLoader
from typing import Any, Coroutine, List, Dict, Tuple
from faker import Faker
import random
from ..evaluation.abstractions.evaluation_metric import EvaluationMetric
from ..evaluation.multi_task_evaluation_service import MultiTaskEvaluationService
from ..modeling.abstractions.model import Model, TInput, TTarget

class MultiTaskEvaluationServiceTestCase(unittest.TestCase):
    def setUp(self):
        fake = Faker()

        self.samples: List[Tuple[str, str]] = [(fake.first_name(), fake.last_name()) for i in range(10)]

        self.model: Model[str, str] = MagicMock(spec=Model)

        self.model.predict_batch = Mock(return_value=[fake.last_name() for i in range(10)])

        self.evaluation_metric_1: EvaluationMetric[str, str, Model[str, str]] = MagicMock(spec=EvaluationMetric)

        self.evaluation_metric_1.calculate_score = Mock(return_value=fake.pyfloat(positive=True))

        self.evaluation_metric_2: EvaluationMetric[str, str, Model[str, str]] = MagicMock(spec=EvaluationMetric)

        self.evaluation_metric_2.calculate_score = Mock(return_value=fake.pyfloat(positive=True))

        self.dataset: Dataset[Tuple[str, str]] = Mock()
        self.dataset.__getitem__ = Mock(return_value=random.choice(self.samples))
        self.dataset.__len__ = Mock(return_value=self.samples.__len__())

        self.event_loop = asyncio.get_event_loop()

        self.evaluation_service: MultiTaskEvaluationService[str, str, Model[str, str]] = MultiTaskEvaluationService[str, str, Model[str, str]](event_loop=self.event_loop)

    def tearDown(self):
        pass

    def test_evaluate_valid_model_metrics_and_dataset_should_return_results_for_each_metric(self):
        evaluation_routine: Coroutine[Any, Any, Dict[str, float]] = self.evaluation_service.evaluate(self.model, self.dataset, 
                    {'metric 1': self.evaluation_metric_1, 'metric 2': self.evaluation_metric_2})

        result: Dict[str, float] = self.event_loop.run_until_complete(evaluation_routine)

        assert len(result.items()) == 2

    def test_evaluation_on_multiple_datasets_valid_model_metrics_and_datasets_should_return_results_for_each_metric_on_each_dataset(self):
        datasets: Dict[str, Dataset[Tuple[str, str]]] = {"set_1": self.dataset, "set_2": self.dataset}

        evaluation_routine: Coroutine[Any, Any, Dict[str, Dict[str, float]]] = self.evaluation_service.evaluate_on_multiple_datasets(self.model, datasets, 
            {'metric 1': self.evaluation_metric_1, 'metric 2': self.evaluation_metric_2})

        result: Dict[str, Dict[str, float]] = self.event_loop.run_until_complete(evaluation_routine)

        assert len(result.items()) == 2


import unittest
from typing import Any, Callable, Coroutine, List, Dict, Tuple, Type

from torch import isin

from ..modeling.abstractions.model import Model
from ..experimentation.default_dict_instance_factory import DefaultDictInstanceFactory
from ..evaluation.abstractions.evaluation_service import EvaluationService
from ..evaluation.multi_task_evaluation_service import MultiTaskEvaluationService
from ..evaluation.abstractions.evaluation_metric import EvaluationMetric
from ..evaluation.custom_evaluation_metric import CustomEvaluationMetric
from ..experimentation.abstractions.machine_learning_experimentation_service import InstanceSettings

CustomEvaluationMetricAlias = CustomEvaluationMetric[float, float, Model[float, float]]

class DefaultDictInstanceFactoryTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_create_multiple_factory_function_available_should_return_valid_instances(self):
        available_types: Dict[str, Type[EvaluationMetric[float, float, Model[float, float]]]] = {'metric_1': lambda **params: CustomEvaluationMetricAlias(lambda x: 1), 'metric_2': lambda **params: CustomEvaluationMetricAlias(lambda x: 0.5)}

        default_instance_factory: DefaultDictInstanceFactory[EvaluationMetric[float, float, Model[float, float]]] = DefaultDictInstanceFactory[EvaluationMetric[float, float, Model[float, float]]](available_types=available_types)

        default_instance_settings: Dict[str, InstanceSettings] = {'metric_1': InstanceSettings(name="metric_1", params={}), 'metric_2': InstanceSettings(name="metric_2", params={})}

        instances: Dict[str, EvaluationMetric[float, float, Model[float, float]]] = default_instance_factory.create(default_instance_settings)

        assert "metric_1" in instances.keys()
        assert "metric_2" in instances.keys()
        assert isinstance(instances["metric_1"], CustomEvaluationMetric)
        assert isinstance(instances["metric_2"], CustomEvaluationMetric)


from abc import ABC, abstractmethod
from typing import Any, Dict, Generic, TypeVar

TSettings = TypeVar('TSettings')
TInstance = TypeVar('TInstance')

class InstanceFactory(Generic[TSettings, TInstance], ABC):

    @abstractmethod
    def create(self, settings: TSettings) -> TInstance:
        pass

    def __call__(self, settings: TSettings) -> TInstance:
        return self.create(settings)


import asyncio
import unittest
from unittest.mock import Mock, patch
from torch.utils.data import Dataset
from typing import Any, Coroutine, List, Dict, Tuple
from faker import Faker
import random

from ..evaluation.abstractions.evaluation_metric import EvaluationContext
from ..parameter_tuning.abstractions.model_factory import ModelFactory
from ..parameter_tuning.abstractions.objective_function import ObjectiveFunction
from ..parameter_tuning.grid_search_tuning_service import GridSearchTuningService
from ..evaluation.multi_task_evaluation_service import MultiTaskEvaluationService
from ..modeling.abstractions.model import Model, TInput, TTarget

class GridSearchTuningServiceTestCase(unittest.TestCase):
    def setUp(self):
        fake = Faker()

        self.samples: List[Tuple[float, float]] = [(fake.pyfloat(positive=True), fake.pyfloat(positive=True)) for i in range(10)]

        self.parameter_tuning_service: GridSearchTuningService[float, float] = GridSearchTuningService[float, float](folds=5)

        self.event_loop = asyncio.get_event_loop()

        self.model_patcher = patch('machine_learning.modeling.abstractions.model.Model', new=Model[float, float])

        model: Model[float, float] = self.model_patcher.start()

        model.predict_batch = Mock(return_value=[fake.pyfloat(positive=True) for i in range(10)])
        model.train = lambda X, y: None
        model.predict = lambda X: fake.pyfloat(positive=True)
        model.__init__ = Mock(return_value=None)
        # model.__class__ = Model[float, float]
        model.__class__.__init__ = lambda x, y: None

        self.model_factory_patcher = patch('machine_learning.parameter_tuning.abstractions.model_factory.ModelFactory')

        self.model_factory: ModelFactory[Model[float, float]] = self.model_factory_patcher.start()
        self.model_factory.create = lambda x: model

        self.dataset_patcher = patch('torch.utils.data.Dataset')

        self.dataset: Dataset[Tuple[float, float]] = self.dataset_patcher.start()
        self.dataset.__getitem__ = lambda x, y: random.choice(self.samples)#Mock(return_value=random.choice([(fake.pyfloat(positive=True), fake.pyfloat(positive=True)) for i in range(10)]))
        self.dataset.__iter__ = lambda x: self.samples.__iter__()#Mock(return_value=[(fake.pyfloat(positive=True), fake.pyfloat(positive=True)) for i in range(10)].__iter__())
        self.dataset.__len__ = lambda : self.samples.__len__()#Mock(return_value=[(fake.pyfloat(positive=True), fake.pyfloat(positive=True)) for i in range(10)].__len__())

        self.objective_function_patcher = patch('machine_learning.parameter_tuning.abstractions.objective_function.ObjectiveFunction')

        self.objective_function: ObjectiveFunction[float, float, Model[float, float]] = self.objective_function_patcher.start()

        self.objective_function.calculate_score = Mock(return_value=fake.pyfloat(positive=True))

    def tearDown(self):
        self.model_patcher.stop()
        self.model_factory_patcher.stop()
        self.dataset_patcher.stop()
        self.objective_function_patcher.stop()

    def test_search_valid_params_should_return_best_params(self):
        search_routine: Coroutine[Any, Any, Dict[str, Any]] = self.parameter_tuning_service.search(model_factory=self.model_factory, params={}, dataset=self.dataset, objective_functions={'test': self.objective_function}, primary_objective='test')

        result: Dict[str, Any] = self.event_loop.run_until_complete(search_routine)


import unittest
from typing import Any, Callable, Coroutine, List, Dict, Tuple, Type

from ..modeling.abstractions.model import Model
from ..experimentation.default_instance_factory import DefaultInstanceFactory
from ..evaluation.abstractions.evaluation_service import EvaluationService
from ..evaluation.multi_task_evaluation_service import MultiTaskEvaluationService
from ..experimentation.abstractions.machine_learning_experimentation_service import InstanceSettings

EvaluationServiceAlias = MultiTaskEvaluationService[float, float, Model[float, float]]

class DefaultInstanceFactoryTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_create_one_type_available_should_return_valid_instance(self):
        available_types: Dict[str, Type[EvaluationService[float, float, Model[float, float]]]] = {'multi_task_evaluation_service': EvaluationServiceAlias}

        default_instance_factory: DefaultInstanceFactory[EvaluationService[float, float, Model[float, float]]] = DefaultInstanceFactory[EvaluationService[float, float, Model[float, float]]](available_types=available_types)

        default_instance_settings: InstanceSettings = InstanceSettings(name="multi_task_evaluation_service", params={"batch_size": 10})

        instance: EvaluationService[float, float, Model[float, float]] = default_instance_factory.create(default_instance_settings)

        assert isinstance(instance, MultiTaskEvaluationService) 

    def test_create_one_factory_function_available_should_return_valid_instance(self):
        available_types: Dict[str, Callable[[InstanceSettings], EvaluationService[float, float, Model[float, float]]]] = {'multi_task_evaluation_service': lambda **params: EvaluationServiceAlias(**params)}

        default_instance_factory: DefaultInstanceFactory[EvaluationService[float, float, Model[float, float]]] = DefaultInstanceFactory[EvaluationService[float, float, Model[float, float]]](available_types=available_types)

        default_instance_settings: InstanceSettings = InstanceSettings(name="multi_task_evaluation_service", params={"batch_size": 10})

        instance: EvaluationService[float, float, Model[float, float]] = default_instance_factory.create(default_instance_settings)

        assert isinstance(instance, MultiTaskEvaluationService) 


from abc import ABC, abstractmethod
from dataclasses import dataclass
import enum
from typing import TypeVar, List, Generic, Dict

from ...modeling.abstractions.model import Model
from ...modeling.abstractions.model import Model, TInput, TTarget
from ...parameter_tuning.abstractions.objective_function import OptimizationType
from ...evaluation.abstractions.evaluation_service import Score

TModel = TypeVar('TModel', bound=Model)

@dataclass
class TrainingContext(Generic[TInput, TTarget, TModel]):
    model: TModel
    dataset_name: str
    current_epoch: int
    current_iteration: int
    scores: Dict[str, List[Score[TInput, TTarget]]]
    _primary_objective: str

    @property
    def primary_scores(self) -> List[Score[TInput, TTarget]]:
        return self.scores[self._primary_objective]

    @property
    def current_scores(self) -> Dict[str, Score[TInput, TTarget]]:
        return {score_name: scores[self.current_epoch - 1] for score_name, scores in self.scores.items() if self.current_epoch > 0}

class StopCondition(Generic[TInput, TTarget, TModel], ABC):
    @abstractmethod
    def reset(self):
        pass

    @abstractmethod
    def is_satisfied(self, context: TrainingContext[TInput, TTarget, TModel]) -> bool:
        pass


from logging import Logger
import logging
from typing import Any, Callable, Coroutine, TypeVar, List, Generic, Optional, Dict, Tuple, Union
from uuid import UUID
import uuid

from attr import asdict
from ..modeling.abstractions.model import Model, TInput, TTarget
from .abstractions.evaluation_metric import EvaluationContext, EvaluationMetric, Prediction, TModel
from .abstractions.evaluation_service import EvaluationService, Score
from .default_evaluation import default_evaluation
import asyncio
import asyncio.tasks
import asyncio.futures
from dataset_handling.dataloader import DataLoader
from torch.utils.data.dataset import Dataset
import nest_asyncio

nest_asyncio.apply()

class MultiTaskEvaluationService(EvaluationService[TInput, TTarget, TModel]):
    def __init__(self, evaluation_hook: Callable[[Logger, EvaluationContext[TInput, TTarget, TModel], TModel, List[TInput], List[TTarget]], List[TTarget]] = default_evaluation, logger: Optional[Logger]=None, batch_size: int = 1, drop_last: bool = True, event_loop: Optional[asyncio.AbstractEventLoop] = None):
        if evaluation_hook is None:
            raise ValueError("evaluation_hook")
        
        self.__logger = logger if not logger is None else logging.getLogger()
        self.__event_loop: asyncio.AbstractEventLoop = event_loop if not event_loop is None else asyncio.get_event_loop()
        self.__batch_size: int = batch_size
        self.__drop_last: bool = drop_last
        self.__evaluation_hook: Callable[[Logger, EvaluationContext[TInput, TTarget, TModel], TModel, List[TInput], List[TTarget]], List[TTarget]] = evaluation_hook

    def __predict_batch(self, evaluation_context: EvaluationContext[TInput, TTarget, TModel], model: TModel, batch: List[Tuple[TInput, TTarget]]) -> List[Prediction]:
        inputs: List[TInput] = [sample[0] for sample in batch]
        targets: List[TInput] = [sample[1] for sample in batch]
        predictions: List[TTarget] = self.__evaluation_hook(self.__logger, evaluation_context, model, inputs, targets)

        combined: List[Tuple[TInput, TTarget, TTarget]] = zip(inputs, predictions, targets)

        return [Prediction(result[0], result[1], result[2]) for result in combined]

    async def evaluate(self, model: TModel, evaluation_dataset: Union[Tuple[str, Dataset[Tuple[TInput, TTarget]]], Dataset[Tuple[TInput, TTarget]]], evaluation_metrics: Dict[str, EvaluationMetric[TInput, TTarget, TModel]], logger: Optional[Logger] = None) -> Dict[str, Score]:
        if logger is None:
            logger = self.__logger
        
        if model is None:
            raise ValueError("model")

        if evaluation_dataset is None:
            raise ValueError("evaluation_dataset")

        if evaluation_metrics is None:
            raise ValueError("evaluation_metrics")

        dataset: Dataset[Tuple[TInput, TTarget]] = None
        dataset_name: str = None

        if isinstance(evaluation_dataset, Tuple):
            dataset = evaluation_dataset[1]
            dataset_name = evaluation_dataset[0]
        else:
            dataset = evaluation_dataset
            dataset_name = type(dataset).__name__

        evaluation_context: EvaluationContext[TInput, TTarget, TModel] = EvaluationContext[TInput, TTarget, TModel](model, dataset_name, [])

        data_loader: DataLoader[Tuple[TInput, TTarget]] = DataLoader[Tuple[TInput, TTarget]](dataset=dataset, batch_size=self.__batch_size, drop_last=self.__drop_last)

        logger.info('Starting evaluation loop...')

        prediction_futures: List[asyncio.Future] = [self.__event_loop.run_in_executor(None, lambda: self.__predict_batch(evaluation_context, model, batch)) for batch in data_loader]
        predictions: List[List[Prediction]] = await asyncio.gather(*prediction_futures, loop=self.__event_loop)

        for t in predictions:
            evaluation_context.predictions.extend(t) 

        result: Dict[str, Score] = {}

        for i, (name, evaluation_metric) in enumerate(evaluation_metrics.items()):
            value: float = evaluation_metric.calculate_score(context=evaluation_context)

            result[name] = Score[TInput, TTarget](value, evaluation_context.predictions, name, dataset_name)

        logger.info('Finished evaluation loop.')
        logger.info({'evaluation_context': evaluation_context})
        logger.info({'scores': result})

        return result

    async def __evaluate(self, model: TModel, evaluation_dataset: Tuple[str, Dataset[Tuple[TInput, TTarget]]], evaluation_metrics: Dict[str, EvaluationMetric[TInput, TTarget, TModel]], logger: Logger) -> Tuple[str, Dict[str, Score]]:
        result = await self.evaluate(model, evaluation_dataset, evaluation_metrics, logger)

        return (evaluation_dataset[0], result)

    async def evaluate_on_multiple_datasets(self, model: TModel, evaluation_datasets: Dict[str, Dataset[Tuple[TInput, TTarget]]], evaluation_metrics: Dict[str, EvaluationMetric[TInput, TTarget, TModel]]) -> Dict[str, Dict[str, Score]]:
        self.__logger.info(f"starting evaluation on {len(evaluation_datasets)} datasets...")
        
        experiment_tasks: List[Coroutine[Any, Any, Tuple[str, Dict[str, Score]]]] = [self.__evaluate(model, dataset, evaluation_metrics, self.__logger.getChild(str(dataset[0]))) for dataset in evaluation_datasets.items()]

        experiment_results: List[Tuple[str, Dict[str, Score]]] = await asyncio.gather(*experiment_tasks, loop=self.__event_loop)

        results = dict(experiment_results)

        self.__logger.info(f"finished evaluation on {len(evaluation_datasets)} datasets.")

        return results


from asyncio import coroutine
import asyncio
from dataclasses import dataclass
from re import S
import unittest
from unittest.mock import AsyncMock, MagicMock, Mock, patch
from torch import isin
from torch.utils.data import Dataset
from dataset_handling.dataloader import DataLoader
from typing import Any, Coroutine, List, Dict, Tuple
from faker import Faker
import random
from ..evaluation.abstractions.evaluation_metric import EvaluationMetric
from ..evaluation.multi_task_evaluation_service import MultiTaskEvaluationService
from ..evaluation.abstractions.evaluation_service import EvaluationService
from ..training.abstractions.training_service import TrainingService
from ..modeling.abstractions.model import Model, TInput, TTarget
from ..training.batch_training_service import BatchTrainingService
from ..training.abstractions.stop_condition import StopCondition
from ..parameter_tuning.abstractions.objective_function import ObjectiveFunction
from ..experimentation.abstractions.machine_learning_experimentation_service import MachineLearningExperimentSettings, InstanceSettings
from ..experimentation.default_machine_learning_experimentation_service import DefaultMachineLearningExperimentationService, MachineLearningExperimentResult, MachineLearningExperimentSettings

class DefaultMachineLearningExperimentationTestCase(unittest.TestCase):
    def setUp(self):
        fake = Faker()

        self.samples: List[Tuple[str, str]] = [(fake.first_name(), fake.last_name()) for i in range(100)]

        self.model: Model[str, str] = AsyncMock(spec=Model)
        self.model.__class__ = AsyncMock

        self.objective_function: ObjectiveFunction[str, str, Model[str, str]] = AsyncMock(spec=ObjectiveFunction)
        self.objective_function.__class__ = AsyncMock

        self.stop_condition: StopCondition[str, str, Model[str, str]] = AsyncMock(spec=StopCondition)
        self.stop_condition.__class__ = AsyncMock

        self.evaluation_metric: EvaluationMetric[str, str, Model[str, str]] = AsyncMock(spec=EvaluationMetric)
        self.stop_condition.__class__ = AsyncMock

        self.evaluation_service: EvaluationService[str, str, Model[str, str]] = AsyncMock(spec=EvaluationService)
        self.evaluation_service.__class__ = AsyncMock

        self.training_service: TrainingService[str, str, Model[str, str]] = AsyncMock(spec=TrainingService)
        self.training_service.__class__ = AsyncMock

        self.dataset: Dataset[Tuple[str, str]] = AsyncMock()
        self.dataset.__class__ = AsyncMock

        self.event_loop = asyncio.get_event_loop()

        self.experimentation_service: DefaultMachineLearningExperimentationService[Model[str, str]] = DefaultMachineLearningExperimentationService[Model[str, str]](logger=None, 
        model_factory=lambda x: self.model, training_service_factory=lambda x: self.training_service, evaluation_service_factory=lambda x: self.evaluation_service, 
        training_dataset_factory=lambda x: self.dataset, test_dataset_factory=lambda x: self.dataset, evaluation_metric_factory=lambda x: self.evaluation_metric, 
        objective_function_factory=lambda x: self.objective_function, stop_condition_factory=lambda x: self.stop_condition, event_loop=self.event_loop)

    def tearDown(self):
        pass

    def test_run_experiment_valid_settings_should_return_experiment_result(self):
        experiment_settings: MachineLearningExperimentSettings = MachineLearningExperimentSettings(name="experiment 1", model_settings=[InstanceSettings(name="", params={})], 
        training_service_settings=[InstanceSettings(name="test 1", params={}), InstanceSettings(name="test 2", params={})], evaluation_service_settings=[InstanceSettings(name="", params={})], training_dataset_settings=[{"": InstanceSettings(name="", params={})}], 
        evaluation_dataset_settings=[{"": InstanceSettings(name="test 1", params={})}, {"": InstanceSettings(name="test 2", params={})}], evaluation_metric_settings=[{"": InstanceSettings(name="", params={})}],
        objective_function_settings=[{"": InstanceSettings(name="", params={})}], stop_condition_settings=[{"": InstanceSettings(name="", params={})}])

        evaluation_routine: Coroutine[Any, Any, MachineLearningExperimentResult[Model[str, str]]] = self.experimentation_service.run_experiment(experiment_settings)

        result: MachineLearningExperimentResult[Model[str, str]] = self.event_loop.run_until_complete(evaluation_routine)

        assert len(result.run_results) == 4

    def test_run_experiments_valid_settings_should_return_experiment_result_for_each_setting(self):
        experiment_settings_1: MachineLearningExperimentSettings = MachineLearningExperimentSettings(name="experiment 1", model_settings=[InstanceSettings(name="", params={})], 
        training_service_settings=[InstanceSettings(name="test 1", params={}), InstanceSettings(name="test 2", params={})], evaluation_service_settings=[InstanceSettings(name="", params={})], training_dataset_settings=[{"": InstanceSettings(name="", params={})}], 
        evaluation_dataset_settings=[{"": InstanceSettings(name="test 1", params={})}, {"": InstanceSettings(name="test 2", params={})}], evaluation_metric_settings=[{"": InstanceSettings(name="", params={})}],
        objective_function_settings=[{"": InstanceSettings(name="", params={})}], stop_condition_settings=[{"": InstanceSettings(name="", params={})}])

        experiment_settings_2: MachineLearningExperimentSettings = MachineLearningExperimentSettings(name="experiment 2", model_settings=[InstanceSettings(name="", params={})], 
        training_service_settings=[InstanceSettings(name="test 1", params={}), InstanceSettings(name="test 2", params={})], evaluation_service_settings=[InstanceSettings(name="", params={})], training_dataset_settings=[{"": InstanceSettings(name="", params={})}], 
        evaluation_dataset_settings=[{"": InstanceSettings(name="test 1", params={})}, {"": InstanceSettings(name="test 2", params={})}], evaluation_metric_settings=[{"": InstanceSettings(name="", params={})}],
        objective_function_settings=[{"": InstanceSettings(name="", params={})}], stop_condition_settings=[{"": InstanceSettings(name="", params={})}])

        evaluation_routine: Coroutine[Any, Any, Dict[str, MachineLearningExperimentResult[Model[str, str]]]] = self.experimentation_service.run_experiments({"ex 1": experiment_settings_1, "ex 2": experiment_settings_2})

        result: Dict[str, MachineLearningExperimentResult[Model[str, str]]] = self.event_loop.run_until_complete(evaluation_routine)

        assert len(list(result.items())) == 2

        for key, item in result.items():
            assert len(item.run_results) == 4


from abc import ABC
from logging import Logger
import logging
from pathlib import Path
from typing import Any, Callable, Coroutine, TypeVar, List, Generic, Optional, Dict, Tuple, Union
import asyncio
import asyncio.tasks
import asyncio.futures
from uuid import UUID
import uuid
from dataset_handling.dataloader import DataLoader
from torch.utils.data import Dataset, random_split
import nest_asyncio
from dataclasses import dataclass

from ..evaluation.default_evaluation import default_evaluation
from ..evaluation.abstractions.evaluation_service import EvaluationService
from ..parameter_tuning.abstractions.objective_function import ObjectiveFunction
from .abstractions.stop_condition import StopCondition, TrainingContext, Score
from ..modeling.abstractions.model import Model, TInput, TTarget
from .abstractions.training_service import TrainingService
from ..evaluation.abstractions.evaluation_metric import EvaluationContext, TModel
from ..evaluation.multi_task_evaluation_service import MultiTaskEvaluationService

nest_asyncio.apply()

@dataclass
class TrainingCheckpoint(Generic[TInput, TTarget]):
    id: UUID
    current_epoch: int
    scores: Dict[str, List[Score[TInput, TTarget]]]

@dataclass
class MultiDatasetTrainingCheckpoint():
    id: UUID
    train_runs: Dict[UUID, str]

class BatchTrainingService(TrainingService[TInput, TTarget, TModel], ABC):
    def __init__(self, train_hook: Callable[[Logger, TrainingContext[TInput, TTarget, TModel], List[TInput], List[TTarget]], None], 
    evaluation_hook: Callable[[Logger, TModel, List[TInput], List[TTarget]], List[TTarget]] = default_evaluation, logger: Optional[Logger]=None, 
    evaluation_service: Optional[EvaluationService[TInput, TTarget, TModel]] = None, 
    batch_size: int = 1, drop_last: bool = True, event_loop: Optional[asyncio.AbstractEventLoop] = None, max_epochs: int = 100, 
    max_iterations: int = 10000, training_dataset_size_ratio: float = 0.8, pre_loop_hook: Optional[Callable[[Logger, TrainingContext[TInput, TTarget, TModel]], None]] = None,
    post_loop_hook: Optional[Callable[[Logger, TrainingContext[TInput, TTarget, TModel]], None]] = None, pre_epoch_hook: Optional[Callable[[Logger, TrainingContext[TInput, TTarget, TModel]], None]] = None, 
    post_epoch_hook: Optional[Callable[[Logger, TrainingContext[TInput, TTarget, TModel], Dataset[Tuple[TInput, TTarget]]], None]] = None, pre_train_hook: Optional[Callable[[Logger, TrainingContext[TInput, TTarget, TModel]], None]] = None,
    post_train_hook: Optional[Callable[[Logger, TrainingContext[TInput, TTarget, TModel]], None]] = None, save_checkpoint_hook: Optional[Callable[[Logger, TrainingCheckpoint[TInput, TTarget]], None]] = None,
    load_checkpoint_hook: Optional[Callable[[Logger, UUID], Optional[TrainingCheckpoint[TInput, TTarget]]]] = None, save_model_hook: Optional[Callable[[Logger, UUID, TModel], None]] = None,
    save_multi_dataset_checkpoint_hook: Optional[Callable[[Logger, MultiDatasetTrainingCheckpoint], None]] = None, load_multi_dataset_checkpoint_hook: Optional[Callable[[Logger, UUID], Optional[MultiDatasetTrainingCheckpoint]]] = None):
        
        if train_hook is None:
            raise ValueError("train_hook")

        self.__logger = logger if not logger is None else logging.getLogger()
        
        if evaluation_service is None:
            evaluation_service = MultiTaskEvaluationService[TInput, TTarget, TModel](logger=self.__logger, batch_size=batch_size, drop_last=drop_last, event_loop=event_loop, evaluation_hook=evaluation_hook)
        
        self.__event_loop: asyncio.AbstractEventLoop = event_loop if not event_loop is None else asyncio.get_event_loop()
        self.__max_epochs: int = max_epochs
        self.__max_iterations: int = max_iterations
        self.__batch_size: int = batch_size
        self.__drop_last: bool = drop_last
        self.__evaluation_service: EvaluationService[TInput, TTarget, TModel, EvaluationContext[TInput, TTarget, TModel]] = evaluation_service
        self.__training_dataset_size_ratio: float = training_dataset_size_ratio

        self.__train_hook: Callable[[Logger, TrainingContext[TInput, TTarget, TModel], List[TInput], List[TTarget]], None] = train_hook
        self.__pre_loop_hook: Optional[Callable[[Logger, TrainingContext[TInput, TTarget, TModel]], None]] = pre_loop_hook
        self.__post_loop_hook: Optional[Callable[[Logger, TrainingContext[TInput, TTarget, TModel]], None]] = post_loop_hook
        self.__pre_epoch_hook: Optional[Callable[[Logger, TrainingContext[TInput, TTarget, TModel]], None]] = pre_epoch_hook
        self.__post_epoch_hook: Optional[Callable[[Logger, TrainingContext[TInput, TTarget, TModel], Dataset[Tuple[TInput, TTarget]]], None]] = post_epoch_hook
        self.__pre_train_hook: Optional[Callable[[Logger, TrainingContext[TInput, TTarget, TModel]], None]] = pre_train_hook
        self.__post_train_hook: Optional[Callable[[Logger, TrainingContext[TInput, TTarget, TModel]], None]] = post_train_hook

        self.__save_checkpoint_hook: Optional[Callable[[Logger, TrainingCheckpoint[TInput, TTarget]], None]] = save_checkpoint_hook
        self.__load_checkpoint_hook: Optional[Callable[[Logger, UUID], Optional[TrainingCheckpoint[TInput, TTarget]]]] = load_checkpoint_hook

        self.__save_multi_dataset_checkpoint_hook: Optional[Callable[[Logger, MultiDatasetTrainingCheckpoint], None]] = save_multi_dataset_checkpoint_hook
        self.__load_multi_dataset_checkpoint_hook: Optional[Callable[[Logger, UUID], Optional[MultiDatasetTrainingCheckpoint]]] = load_multi_dataset_checkpoint_hook

        self.__save_model_hook: Optional[Callable[[Logger, UUID, TModel], None]] = save_model_hook

    def __is_any_stop_condition_satisfied(self, training_context: TrainingContext[TInput, TTarget, TModel], stop_conditions: Dict[str, StopCondition[TInput, TTarget, TModel]], logger: Optional[Logger] = None) -> bool:
        if logger is None:
            logger = self.__logger
        
        self.__logger.info("Checking stop conditions...")
        is_any_satisfied: bool = False

        if training_context.current_epoch > self.__max_epochs: 
            self.__logger.info("Max number of epochs reached.")
            is_any_satisfied = True
        elif training_context.current_iteration > self.__max_iterations:
            self.__logger.info("Max number of iterations reached.")
        else:
            for key, condition in stop_conditions.items():
                is_any_satisfied |= condition.is_satisfied(training_context)

                if(is_any_satisfied):
                    self.__logger.info('Condition named "{key}" is satisfied'.format(key=key))
                    break

        self.__logger.info("Finished checking stop conditions.")
        return is_any_satisfied

    def __load_checkpoint(self, logger: Logger, id: UUID) -> Optional[TrainingCheckpoint]:
        if self.__load_checkpoint_hook is None:
            return None

        logger.info("Loading last checkpoint...")

        return self.__load_checkpoint_hook(logger, id)

    def __save_checkpoint(self, logger: Logger, checkpoint: TrainingCheckpoint):
        if not self.__save_checkpoint_hook is None:
            logger.info("creating checkpoint...")

            self.__save_checkpoint_hook(logger, checkpoint)

            logger.info("checkpoint created.")

    def __load_multi_dataset_checkpoint(self, logger: Logger, id: UUID) -> Optional[MultiDatasetTrainingCheckpoint]:
        if self.__load_multi_dataset_checkpoint_hook is None:
            return None

        logger.info("Loading last checkpoint...")

        return self.__load_multi_dataset_checkpoint_hook(logger, id)

    def __save_multi_dataset_checkpoint(self, logger: Logger, checkpoint: MultiDatasetTrainingCheckpoint):
        if not self.__save_multi_dataset_checkpoint_hook is None:
            logger.info("creating checkpoint...")

            self.__save_multi_dataset_checkpoint_hook(logger, checkpoint)

            logger.info("checkpoint created.")

    def __save_model(self, logger: Logger, id: UUID, model: TModel):
        if not self.__save_model_hook is None:
            logger.info("saving model...")

            self.__save_model_hook(logger, id, model)

            logger.info("model saved.")

    async def __execute_train_loop(self, id: UUID, model: TModel, training_context: TrainingContext[TInput, TTarget, TModel], stop_conditions: Dict[str, StopCondition[TInput, TTarget, TModel]], objective_functions: Dict[str, ObjectiveFunction[TInput, TTarget, TModel]], primary_objective: str, logger: Logger, training_dataset: Tuple[str, Dataset[Tuple[TInput, TTarget]]], validation_dataset: Tuple[str, Dataset[Tuple[TInput, TTarget]]]):
        training_data_loader: DataLoader[Tuple[TInput, TTarget]] = DataLoader[Tuple[TInput, TTarget]](dataset=training_dataset[1], batch_size=self.__batch_size, drop_last=self.__drop_last)
        
        while not self.__is_any_stop_condition_satisfied(training_context=training_context, stop_conditions=stop_conditions):
            training_context.current_epoch += 1

            if not self.__pre_epoch_hook is None:
                logger.debug("Executing pre epoch hook.")
                self.__pre_train_hook(logger, training_context)

            for batch_index, batch in enumerate(training_data_loader):
                training_context.current_iteration += 1

                if not self.__pre_train_hook is None:
                    logger.debug("Executing pre training hook.")
                    self.__pre_train_hook(logger, training_context)

                inputs: List[TInput] = [value[0] for value in batch]
                targets: List[TTarget] = [value[1] for value in batch]

                logger.debug("Executing training hook.")
                self.__train_hook(logger, training_context, inputs, targets)

                if not self.__post_train_hook is None:
                    logger.debug("Executing post training hook.")
                    self.__post_train_hook(logger, training_context)

            logger.info("Evaluating current model.")
            evaluation_scores: Dict[str, Score] = await self.__evaluation_service.evaluate(model=model, evaluation_dataset=validation_dataset, evaluation_metrics=objective_functions)
            logger.info("finished evaluating current model.")

            for key, evaluation_score in evaluation_scores.items():
                training_context.scores[key].append(evaluation_score)

            if not self.__post_epoch_hook is None:
                logger.debug("Executing post epoch hook.")
                self.__post_epoch_hook(logger, training_context, validation_dataset[1])

            logger.info({'training_context': training_context})

            self.__save_model(logger, id, training_context.model)

            new_checkpoint: TrainingCheckpoint[TInput, TTarget] = TrainingCheckpoint[TInput, TTarget](id=id, current_epoch=training_context.current_epoch, scores=training_context.scores)
            self.__save_checkpoint(logger, new_checkpoint)

        logger.info("Finished training loop.")

        if not self.__post_loop_hook is None:
            logger.debug("Executing post loop hook.")
            self.__post_loop_hook(logger, training_context)

        return model

    async def train(self, model: TModel, dataset: Union[Tuple[str, Dataset[Tuple[TInput, TTarget]]], Dataset[Tuple[TInput, TTarget]]], stop_conditions: Dict[str, StopCondition[TInput, TTarget, TModel]], objective_functions: Dict[str, ObjectiveFunction[TInput, TTarget, TModel]], primary_objective: Optional[str] = None, validation_dataset: Optional[Union[Tuple[str, Dataset[Tuple[TInput, TTarget]]], Dataset[Tuple[TInput, TTarget]]]] = None, logger: Optional[Logger] = None, id: Optional[UUID] = None) -> TModel:
        logger = logger if not logger is None else self.__logger
        id = id if not id is None else uuid.uuid4()
        
        if model is None:
            raise ValueError("model")

        if dataset is None:
            raise ValueError("dataset")

        if stop_conditions is None:
            raise ValueError("stop_conditions")

        if objective_functions is None:
            raise ValueError("objective_functions can't be empty")

        if primary_objective is None:
            primary_objective = list(objective_functions.keys())[0]

        checkpoint: Optional[TrainingCheckpoint[TInput, TTarget]] = self.__load_checkpoint(logger, id)

        current_dataset: Tuple[str, Dataset[Tuple[TInput, TTarget]]] = None
        training_context: TrainingContext[TInput, TTarget, TModel] = None
        training_dataset: Tuple[str, Dataset[Tuple[TInput, TTarget]]] = None

        if isinstance(dataset, Tuple):
            current_dataset = dataset
        else:
            current_dataset = (type(dataset).__name__, dataset)

        training_context = TrainingContext[TInput, TTarget, TModel](model=model, dataset_name=current_dataset[0], scores={objective: [] for objective in objective_functions.keys()}, _primary_objective=primary_objective, current_epoch=0, current_iteration=0)

        training_size: int = int(len(current_dataset[1]) * self.__training_dataset_size_ratio)
        validation_size: int = int(len(current_dataset[1]) - training_size)

        if validation_dataset is None: 
            training_split, validation_split = random_split(current_dataset[1], [training_size, validation_size])

            training_dataset = (current_dataset[0], training_split)
            validation_dataset = (current_dataset[0], validation_split)
        else:
            training_dataset = current_dataset

        if not checkpoint is None:
            return await self.continue_training(model, stop_conditions, objective_functions, primary_objective, checkpoint, training_dataset, validation_dataset, logger)

        logger.info('Starting training loop...')

        if not self.__pre_loop_hook is None:
            logger.debug("Executing pre loop hook.")
            self.__pre_loop_hook(logger, training_context)

        new_checkpoint: TrainingCheckpoint[TInput, TTarget] = TrainingCheckpoint[TInput, TTarget](id=id, current_epoch=training_context.current_epoch, scores=training_context.scores)
        self.__save_checkpoint(logger, new_checkpoint)

        return await self.__execute_train_loop(id, model, training_context, stop_conditions, objective_functions, primary_objective, logger, training_dataset, validation_dataset)

    async def continue_training(self, model: TModel, stop_conditions: Dict[str, StopCondition[TInput, TTarget, TModel]], objective_functions: Dict[str, ObjectiveFunction[TInput, TTarget, TModel]], primary_objective: str, id_or_checkpoint: Union[UUID, TrainingCheckpoint], training_dataset: Tuple[Dataset[Tuple[TInput, TTarget]]], validation_dataset: Tuple[Dataset[Tuple[TInput, TTarget]]], logger: Optional[Logger] = None) -> TModel:
        if logger is None:
            logger = self.__logger

        checkpoint: TrainingCheckpoint = None

        if isinstance(id_or_checkpoint, UUID):
            checkpoint = self.__load_checkpoint(self.__logger, id_or_checkpoint)
        elif isinstance(id_or_checkpoint, TrainingCheckpoint):
            checkpoint = id_or_checkpoint

        if checkpoint is None:
            raise ValueError("checkpoint")

        model: TModel = model
        training_context: TrainingContext[TInput, TTarget, TModel] = TrainingContext[TInput, TTarget, TModel](model=model, dataset_name=training_dataset[0], scores=checkpoint.scores, _primary_objective=primary_objective, current_epoch=checkpoint.current_epoch, current_iteration=0)
        stop_conditions: Dict[str, StopCondition[TInput, TTarget, TModel]] = stop_conditions
        objective_functions: Dict[str, ObjectiveFunction[TInput, TTarget, TModel]] = objective_functions
        primary_objective: str = primary_objective
        id: UUID = checkpoint.id

        logger.info("Continuing training loop from last checkpoint...")

        return await self.__execute_train_loop(id, model, training_context, stop_conditions, objective_functions, primary_objective, logger, training_dataset, validation_dataset)

    async def __execute_multi_train_loop(self, model: TModel, train_runs: Dict[UUID, Tuple[str, Dataset[Tuple[TInput, TTarget]]]], stop_conditions: Dict[str, StopCondition[TInput, TTarget, TModel]], objective_functions: Dict[str, ObjectiveFunction[TInput, TTarget, TModel]], primary_objective: str, validation_dataset: Optional[Union[Tuple[str, Dataset[Tuple[TInput, TTarget]]], Dataset[Tuple[TInput, TTarget]]]] = None) -> TModel:
        for train_run_id, dataset in train_runs.items():
            training_run_logger: Logger = self.__logger.getChild(str(dataset[0]))
            model = await self.train(model, dataset, stop_conditions, objective_functions, primary_objective, validation_dataset, training_run_logger, train_run_id)

        self.__logger.info(f"finished training on {len(train_runs.items())} datasets.")

        return model

    async def train_on_multiple_datasets(self, model: TModel, datasets: Dict[str, Dataset[Tuple[TInput, TTarget]]], stop_conditions: Dict[str, StopCondition[TInput, TTarget, TModel]], objective_functions: Dict[str, ObjectiveFunction[TInput, TTarget, TModel]], primary_objective: Optional[str] = None, validation_dataset: Optional[Union[Tuple[str, Dataset[Tuple[TInput, TTarget]]], Dataset[Tuple[TInput, TTarget]]]] = None, id: Optional[UUID] = None) -> TModel:

        id = id if not id is None else uuid.uuid4()

        checkpoint: Optional[MultiDatasetTrainingCheckpoint] = self.__load_multi_dataset_checkpoint(self.__logger, id)

        if not checkpoint is None:
            return await self.continue_multi_dataset_training(model, stop_conditions, objective_functions, primary_objective, checkpoint, datasets, validation_dataset)

        self.__logger.info(f"starting training on {len(datasets)} datasets...")

        train_runs: Dict[UUID, Tuple[str, Dataset[Tuple[TInput, TTarget]]]] = {uuid.uuid4(): dataset for dataset in datasets.items()}

        new_checkpoint: MultiDatasetTrainingCheckpoint = MultiDatasetTrainingCheckpoint(id, {train_run_id: dataset[0] for train_run_id, dataset in train_runs.items()})
        self.__save_multi_dataset_checkpoint(self.__logger, new_checkpoint)

        return await self.__execute_multi_train_loop(model, train_runs, stop_conditions, objective_functions, primary_objective, validation_dataset)

    async def continue_multi_dataset_training(self, model: TModel, stop_conditions: Dict[str, StopCondition[TInput, TTarget, TModel]], objective_functions: Dict[str, ObjectiveFunction[TInput, TTarget, TModel]], primary_objective: str, id_or_checkpoint: Union[UUID, MultiDatasetTrainingCheckpoint], datasets: Dict[str, Dataset[Tuple[TInput, TTarget]]], validation_dataset: Optional[Union[Tuple[str, Dataset[Tuple[TInput, TTarget]]], Dataset[Tuple[TInput, TTarget]]]]) -> TModel:
        checkpoint: MultiDatasetTrainingCheckpoint = None

        if isinstance(id_or_checkpoint, UUID):
            checkpoint = self.__load_multi_dataset_checkpoint(self.__logger, id_or_checkpoint)
        elif isinstance(id_or_checkpoint, MultiDatasetTrainingCheckpoint):
            checkpoint = id_or_checkpoint

        if checkpoint is None:
            raise ValueError("checkpoint")

        model: TModel = model
        train_runs: Dict[UUID, Tuple[str, Dataset[Tuple[TInput, TTarget]]]] = {train_run_id: datasets[dataset_name] for train_run_id, dataset_name in checkpoint.train_runs.items()}
        stop_conditions: Dict[str, StopCondition[TInput, TTarget, TModel]] = stop_conditions
        objective_functions: Dict[str, ObjectiveFunction[TInput, TTarget, TModel]] = objective_functions
        primary_objective: str = primary_objective

        self.__logger.info("Continuing training loop from last checkpoint...")

        return await self.__execute_multi_train_loop(model, train_runs, stop_conditions, objective_functions, primary_objective, validation_dataset)



from asyncio import coroutine
import asyncio
from dataclasses import dataclass
from re import S
import unittest
from unittest.mock import MagicMock, Mock, patch
from torch import isin
from torch.utils.data import Dataset
from dataset_handling.dataloader import DataLoader
from typing import Any, Coroutine, List, Dict, Tuple
from faker import Faker
import random
from ..evaluation.abstractions.evaluation_metric import EvaluationMetric
from ..evaluation.multi_task_evaluation_service import MultiTaskEvaluationService
from ..modeling.abstractions.model import Model, TInput, TTarget
from ..training.batch_training_service import BatchTrainingService
from ..training.abstractions.stop_condition import StopCondition
from ..parameter_tuning.abstractions.objective_function import ObjectiveFunction

class BatchTrainingServiceTestCase(unittest.TestCase):
    def setUp(self):
        fake = Faker()

        self.samples: List[Tuple[str, str]] = [(fake.first_name(), fake.last_name()) for i in range(100)]

        self.model: Model[str, str] = MagicMock(spec=Model)
        self.model.predict_batch = Mock(return_value=[fake.last_name() for i in range(10)])

        self.objective_function_1: ObjectiveFunction[str, str, Model[str, str]] = MagicMock(spec=ObjectiveFunction)
        self.objective_function_1.calculate_score = Mock(return_value=fake.pyfloat(positive=True))

        self.objective_function_2: ObjectiveFunction[str, str, Model[str, str]] = MagicMock(spec=ObjectiveFunction)
        self.objective_function_2.calculate_score = Mock(return_value=fake.pyfloat(positive=True))

        self.dataset: Dataset[Tuple[str, str]] = Mock()
        self.dataset.__getitem__ = Mock(return_value=random.choice(self.samples))
        self.dataset.__len__ = Mock(return_value=self.samples.__len__())

        self.event_loop = asyncio.get_event_loop()

        self.training_service: BatchTrainingService[str, str, Model[str, str]] = BatchTrainingService[str, str, Model[str, str]](train_hook=lambda w, x, y, z: x, event_loop=self.event_loop)

    def tearDown(self):
        pass

    def test_train_valid_objectives_and_dataset_should_return_trained_model(self):
        evaluation_routine: Coroutine[Any, Any, Model[str, str]] = self.training_service.train(self.model, self.dataset, {}, 
                    {'objective 1': self.objective_function_1, 'objective 2': self.objective_function_2}, None)

        trained_model: Model[str, str] = self.event_loop.run_until_complete(evaluation_routine)

    def test_train_on_multiple_datasets_valid_objectives_and_datasets_should_return_trained_model(self):
        datasets: Dict[str, Dataset[Tuple[str, str]]] = {"set_1": self.dataset, "set_2": self.dataset}

        evaluation_routine: Coroutine[Any, Any, Model[str, str]] = self.training_service.train_on_multiple_datasets(self.model, datasets, {},
            {'objective 1': self.objective_function_1, 'objective 2': self.objective_function_2}, None)

        trained_model: Model[str, str] = self.event_loop.run_until_complete(evaluation_routine)